package Socket_IFC;

// ================================================================
// Copyright (c) Bluespec, Inc., 2007 All Rights Reserved

// ----------------
// Imports from BSV libs

import EdgeFIFOs::*;

// ----------------
// Requests across a socket

typedef enum { NOP, WR, RD } ReqOp
        deriving (Bits, Eq);

function Action display_ReqOp (ReqOp op);
   case (op)
      NOP : $write ("NOP");
      WR  : $write ("WR");
      RD  : $write ("RD");
   endcase
endfunction

typedef Bit#(10)  ReqInfo;
typedef Bit#(32)  ReqAddr;
typedef Bit#(32)  ReqData;

typedef struct {
  ReqOp                  reqOp;
  ReqInfo                reqInfo;
  ReqAddr                reqAddr;
  ReqData                reqData;
} Socket_Req
  deriving (Bits);

// this utility function will help display the requests  
function Action display_Socket_Req (Socket_Req req);
   action
     $write ("Socket_Req{");
     display_ReqOp (req.reqOp);
     if (req.reqOp != NOP)
        $write (", %h, %h, %h}", req.reqInfo, req.reqAddr, req.reqData);
     $display ("}");
   endaction
endfunction

// ================================================================
// Socket Responses

typedef enum { NOP, OK } RespOp
        deriving (Bits, Eq);

// this utility function will help display the responses  
function Action display_RespOp (RespOp op);
   case (op)
      NOP  : $write ("NOP");
      OK   : $write ("OK");
   endcase
endfunction

typedef Bit#(10)   RespInfo;
typedef Bit#(1)    RespAddr;
typedef Bit#(32)   RespData;

typedef struct {
  RespOp                 respOp;
  RespInfo               respInfo;
  RespAddr               respAddr;
  RespData               respData;
} Socket_Resp
  deriving (Eq, Bits);

function Action display_Socket_Resp (Socket_Resp resp);
   action
     $write ("Socket_Resp{");
     display_RespOp (resp.respOp);
     if (resp.respOp != NOP)
        $write (", %h, %h, %h", resp.respInfo, resp.respAddr, resp.respData);
     $display ("}");
   endaction
endfunction

// ================================================================
// Generic Socket Interfaces
// These interfaces will be "always ready, always enabled"
// so that there are no extra signal wires beyond what is implied by
// the argument and result types of the methods

// ----------------
// A master interface
//   exists on an Initiator    (facing a slave interface on an Interconnect)
//      and on an Interconnect (facing a slave interface on a Target)

interface Socket_master_req_ifc;
  method Socket_Req   getReq      ();
  method Action       reqAccept   (Bool  accept);
endinterface: Socket_master_req_ifc

interface Socket_master_resp_ifc;
  method Action      putResp    (Socket_Resp  resp);
  method Bool        respAccept ();
endinterface: Socket_master_resp_ifc

interface Socket_master_ifc;
  interface Socket_master_req_ifc    master_req;
  interface Socket_master_resp_ifc   master_resp;
endinterface: Socket_master_ifc

// The following function produces a Socket master interface from the
// interfaces of two fifos containing requests and responses

function Socket_master_ifc
         fifos_to_master_ifc (FIFOF#(Socket_Req)  reqs,
                              FIFOF#(Socket_Resp) resps);
  return
     (interface Socket_master_ifc
        // Request sub-interface
        interface Socket_master_req_ifc  master_req;
          method Socket_Req  getReq  ();
            let req       = reqs.first ();
            let idleReq   = req ();
            idleReq.reqOp = NOP;
            return (reqs.notEmpty () ? req : idleReq);
          endmethod

          method Action  reqAccept  (Bool accept);
            if (accept && reqs.notEmpty ())
              reqs.deq ();
          endmethod
        endinterface

        // Response sub-interface
        interface Socket_master_resp_ifc  master_resp;
          method Action  putResp (Socket_Resp  resp);
            if ((resp.respOp != NOP) && resps.notFull ())
              resps.enq (resp);
          endmethod

          method Bool  respAccept ();
            return resps.notFull();
          endmethod
        endinterface
      endinterface);
endfunction: fifos_to_master_ifc

// ----------------
// A slave interface
//   exists on an Interconnect (facing a master interface on an Initiator)
//      and on a  Target       (facing a master interface on an Interconnect)

interface Socket_slave_req_ifc;
  method Action       putReq     (Socket_Req req);
  method Bool         reqAccept  ();
endinterface: Socket_slave_req_ifc

interface Socket_slave_resp_ifc;
  method Socket_Resp  getResp    ();
  method Action       respAccept (Bool  accept);
endinterface: Socket_slave_resp_ifc

interface Socket_slave_ifc;
  interface Socket_slave_req_ifc   slave_req;
  interface Socket_slave_resp_ifc  slave_resp;
endinterface: Socket_slave_ifc

// The following function produces a Socket slave interface from the
// interfaces of two fifos containing requests and responses

function Socket_slave_ifc
         fifos_to_slave_ifc (FIFOF#(Socket_Req)  reqs,
                             FIFOF#(Socket_Resp) resps);
  return
     (interface Socket_slave_ifc
        // Request sub-interface
        interface Socket_slave_req_ifc  slave_req;
          method Action  putReq  (Socket_Req  req);
            if ((req.reqOp != NOP) && (reqs.notFull()))
              reqs.enq (req);
          endmethod

          method Bool  reqAccept  ();
            return reqs.notFull();
          endmethod
        endinterface

        // Response sub-interface
        interface Socket_slave_resp_ifc  slave_resp;
          method Socket_Resp  getResp ();
            let resp        = resps.first ();
            let nullResp    = resp;
            nullResp.respOp = NOP;
            return (resps.notEmpty () ? resp : nullResp);
          endmethod

          method Action  respAccept (Bool  accept);
            if (accept && resps.notEmpty ())
              resps.deq();
          endmethod
        endinterface
      endinterface);
endfunction: fifos_to_slave_ifc

// ================================================================

endpackage: Socket_IFC
