package Switch;

// ================================================================
// Copyright (c) Bluespec, Inc., 2007 All Rights Reserved

// ----------------
// imports

import Socket_IFC::*;

import EdgeFIFOs::*;

// ----------------
// Interface for the Switch

interface Switch_IFC;
   // we have 2 slave interfaces towards the initiators
   interface Socket_slave_ifc    initiator_0;
   interface Socket_slave_ifc    initiator_1;

   // we have 2 master interfaces towards the targets
   interface Socket_master_ifc   target_0;
   interface Socket_master_ifc   target_1;
endinterface: Switch_IFC

// ================================================================
// Address-decode functions

function Bool  addr_is_for_target_0 (ReqAddr addr);
   return (addr[0] == 0);
endfunction

function Bool  addr_is_for_target_1 (ReqAddr addr);
   return (addr[0] == 1);
endfunction

function Bool  addr_is_for_initiator_0 (RespAddr addr);
   return (addr == 0);
endfunction

function Bool  addr_is_for_initiator_1 (RespAddr addr);
   return (addr == 1);
endfunction

// ================================================================
// The Switch Module

(* synthesize, always_ready, always_enabled *)
module mkSwitch (Switch_IFC);

   // ----------------------------------------------------------------
   // INTERNAL STATE

   // INITIATOR FIFOs

   // ---- for incoming requests (from initiators)
   FIFOF#(Socket_Req)        from_initiator_0        <- mkPipelineFIFO_ug_enq;
   FIFOF#(Socket_Req)        from_initiator_1        <- mkPipelineFIFO_ug_enq;

   // ---- FIFOs for outgoing resps (to initiators)
   FIFOF#(Socket_Resp)       to_initiator_0          <- mkPipelineFIFO_ug_deq;
   FIFOF#(Socket_Resp)       to_initiator_1          <- mkPipelineFIFO_ug_deq;
  
   // TARGET FIFOs

   // ---- FIFOs for outgoing requests (to targets)
   FIFOF#(Socket_Req)        to_target_0             <- mkBypassFIFO_ug_deq;
   FIFOF#(Socket_Req)        to_target_1             <- mkBypassFIFO_ug_deq;

   // ---- FIFOs for incoming resps (from targets)
   FIFOF#(Socket_Resp)       from_target_0           <- mkBypassFIFO_ug_enq;
   FIFOF#(Socket_Resp)       from_target_1           <- mkBypassFIFO_ug_enq;

   // ---- Additional state for round-robin arbitration
   (* doc = " When this register is TRUE, requsts from " *)
   (* doc = " initiator 0 will go to target 0 " *)
   Reg#(Bool) to_target_0_round_robin    <- mkReg (False);

   (* doc = " When this register is TRUE, requsts from " *)
   (* doc = " initiator 0 will go to target 1 " *)
   Reg#(Bool) to_target_1_round_robin    <- mkReg (False);

   (* doc = " When this register is TRUE, responses from " *)
   (* doc = " target 0 will go to initiator 0 " *)
   Reg#(Bool) to_initiator_0_round_robin <- mkReg (False);

   (* doc = " When this register is TRUE, responses from " *)
   (* doc = " target 0 will go to initiator 1 " *)
   Reg#(Bool) to_initiator_1_round_robin <- mkReg (False);

   // ================================================================
   // RULES (the interconnect functionality)

   // ----------------------------------------------------------------
   // Requests from initiators to targets

   let req_from_initiator_0 = from_initiator_0.first();
  
   let req_from_initiator_1 = from_initiator_1.first();
  
   // ----------------
   // To Target 0

   // We'll give priority to the case where both initiators are fighting over target 0,
   // and then allow the individual cases
   (* descending_urgency = "both_initiators_to_target_0, initiator_0_to_target_0, initiator_1_to_target_0" *)

   // Both initiators to Target 0
   (* doc = "This rule will fire whenever both initiators are" *)
   (* doc = "trying to send a request to target 0" *)
   rule both_initiators_to_target_0
                   (addr_is_for_target_0 (req_from_initiator_0.reqAddr) &&
                    addr_is_for_target_0 (req_from_initiator_1.reqAddr));
     if (to_target_0_round_robin) begin
        to_target_0.enq (req_from_initiator_0);
        from_initiator_0.deq();
     end
     else begin
        to_target_0.enq (req_from_initiator_1);
        from_initiator_1.deq();
     end
     to_target_0_round_robin <= ! to_target_0_round_robin;
   endrule

   // Initiator 0 to Target 0
   (* doc = "This rule will fire whenever initiator 0 is" *)
   (* doc = "trying to send a request to target 0" *)
   rule initiator_0_to_target_0
                   (addr_is_for_target_0 (req_from_initiator_0.reqAddr));
     to_target_0.enq (req_from_initiator_0);
     from_initiator_0.deq();
   endrule

   // Initiator 1 to Target 0
   (* doc = "This rule will fire whenever initiator 1 is" *)
   (* doc = "trying to send a request to target 0" *)
   rule initiator_1_to_target_0
                   (addr_is_for_target_0 (req_from_initiator_1.reqAddr));
     to_target_0.enq (req_from_initiator_1);
     from_initiator_1.deq();
   endrule

   // ----------------
   // To Target 1

   // We'll give priority to the case where both initiators are fighting over target 1,
   // and then allow the individual cases
   (* descending_urgency = "both_initiators_to_target_1, initiator_0_to_target_1, initiator_1_to_target_1" *)

   // Both initiators to Target 1
   (* doc = "This rule will fire whenever both initiators are" *)
   (* doc = "trying to send a request to target 1" *)
   rule both_initiators_to_target_1
                   (addr_is_for_target_1 (req_from_initiator_0.reqAddr) &&
                    addr_is_for_target_1 (req_from_initiator_1.reqAddr));
     if (to_target_1_round_robin) begin
        to_target_1.enq (req_from_initiator_0);
        from_initiator_0.deq();
     end
     else begin
        to_target_1.enq (req_from_initiator_1);
        from_initiator_1.deq();
     end
     to_target_1_round_robin <= ! to_target_1_round_robin;
   endrule

   // Initiator 0 to Target 1
   (* doc = "This rule will fire whenever initiator 0 is" *)
   (* doc = "trying to send a request to target 1" *)
   rule initiator_0_to_target_1
                   (addr_is_for_target_1 (req_from_initiator_0.reqAddr));
      to_target_1.enq(req_from_initiator_0);
      from_initiator_0.deq();
   endrule

   // Initiator 1 to Target 1
   (* doc = "This rule will fire whenever initiator 1 is" *)
   (* doc = "trying to send a request to target 1" *)
   rule initiator_1_to_target_1
                   (addr_is_for_target_1 (req_from_initiator_1.reqAddr));
      to_target_1.enq(req_from_initiator_1);
      from_initiator_1.deq();
   endrule

   // ----------------------------------------------------------------
   // Responses from targets to initiators

   let resp_from_target_0 = from_target_0.first();

   let resp_from_target_1 = from_target_1.first();

   // ----------------
   // To Initiator 0

   // We'll give priority to the case where both targets are trying to send
   // a response to initiator 0, and then allow the individual cases
   (* descending_urgency = "both_targets_to_initiator_0, target_0_to_initiator_0, target_1_to_initiator_0" *)

   // Both targets to Initiator 0
   (* doc = "This rule will fire whenever both targets are" *)
   (* doc = "trying to send a response to initiator 0" *)
   rule both_targets_to_initiator_0
                   (addr_is_for_initiator_0 (resp_from_target_0.respAddr) &&
                    addr_is_for_initiator_0 (resp_from_target_1.respAddr));
     if (to_initiator_0_round_robin) begin
        to_initiator_0.enq (resp_from_target_0);
        from_target_0.deq();
     end
     else begin
        to_initiator_0.enq (resp_from_target_1);
        from_target_1.deq();
     end
     to_initiator_0_round_robin <= ! to_initiator_0_round_robin;
   endrule

   // Target 0 to Initiator 0
   (* doc = "This rule will fire whenever target 0 is" *)
   (* doc = "trying to send a response to initiator 0" *)
   rule target_0_to_initiator_0
                   (addr_is_for_initiator_0 (resp_from_target_0.respAddr));
     to_initiator_0.enq (resp_from_target_0);
     from_target_0.deq();
   endrule

   // Target 1 to Initiator 0
   (* doc = "This rule will fire whenever target 1 is" *)
   (* doc = "trying to send a response to initiator 0" *)
   rule target_1_to_initiator_0
                   (addr_is_for_initiator_0 (resp_from_target_1.respAddr));
     to_initiator_0.enq (resp_from_target_1);
     from_target_1.deq();
   endrule

   // ----------------
   // To Initiator 1

   // We'll give priority to the case where both targets are trying to send
   // a response to initiator 1, and then allow the individual cases
   (* descending_urgency = "both_targets_to_initiator_1, target_0_to_initiator_1, target_1_to_initiator_1" *)

   // Both targets to Initiator 1
   (* doc = "This rule will fire whenever both targets are" *)
   (* doc = "trying to send a response to initiator 1" *)
   rule both_targets_to_initiator_1
                   (addr_is_for_initiator_1 (resp_from_target_0.respAddr) &&
                    addr_is_for_initiator_1 (resp_from_target_1.respAddr));
     if (to_initiator_1_round_robin) begin
        to_initiator_1.enq (resp_from_target_0);
        from_target_0.deq();
     end
     else begin
        to_initiator_1.enq (resp_from_target_1);
        from_target_1.deq();
     end
     to_initiator_1_round_robin <= ! to_initiator_1_round_robin;
   endrule

   // Target 0 to Initiator 1
   (* doc = "This rule will fire whenever target 0 is" *)
   (* doc = "trying to send a response to initiator 1" *)
   rule target_0_to_initiator_1
                   (addr_is_for_initiator_1 (resp_from_target_0.respAddr));
     to_initiator_1.enq (resp_from_target_0);
     from_target_0.deq();
   endrule

   // Target 1 to Initiator 1
   (* doc = "This rule will fire whenever target 1 is" *)
   (* doc = "trying to send a response to initiator 1" *)
   rule target_1_to_initiator_1
                   (addr_is_for_initiator_1 (resp_from_target_1.respAddr));
     to_initiator_1.enq (resp_from_target_1);
     from_target_1.deq();
   endrule

   // ================================================================
   // INTERFACES

   // ---- Initiators
   interface initiator_0 = fifos_to_slave_ifc (from_initiator_0, to_initiator_0);
   interface initiator_1 = fifos_to_slave_ifc (from_initiator_1, to_initiator_1);

   // ---- Targets
   interface target_0    = fifos_to_master_ifc (to_target_0, from_target_0);
   interface target_1    = fifos_to_master_ifc (to_target_1, from_target_1);

endmodule: mkSwitch

endpackage: Switch
