package Switch;

// ================================================================
// Copyright (c) Bluespec, Inc., 2007 All Rights Reserved

// ----------------
// Imports from BSV libs

import FIFO::*;
import GetPut::*;
import ClientServer::*;

// ----------------
// Other imports

import Socket_IFC::*;

// ----------------
// Interface for the Switch

interface Switch_IFC;
   interface Socket_slave_ifc    initiator_0;
   interface Socket_slave_ifc    initiator_1;

   interface Socket_master_ifc   target_0;
   interface Socket_master_ifc   target_1;
endinterface: Switch_IFC

// ================================================================
// Address-decode functions

function Bool  addr_is_for_target_0 (ReqAddr addr);
   return (addr[0] == 0);
endfunction

function Bool  addr_is_for_target_1 (ReqAddr addr);
   return (addr[0] == 1);
endfunction

function Bool  addr_is_for_initiator_0 (RespAddr addr);
   return (addr == 0);
endfunction

function Bool  addr_is_for_initiator_1 (RespAddr addr);
   return (addr == 1);
endfunction

// ================================================================
// The Switch Module

(* synthesize *)
module mkSwitch (Switch_IFC);

   // ----------------------------------------------------------------
   // INTERNAL STATE

   // INITIATOR FIFOs

   // ---- for incoming requests (from initiators)
   FIFO#(Socket_Req)        from_initiator_0        <- mkFIFO;
   FIFO#(Socket_Req)        from_initiator_1        <- mkFIFO;

   // ---- FIFOs for outgoing resps (to initiators)
   FIFO#(Socket_Resp)       to_initiator_0          <- mkFIFO;
   FIFO#(Socket_Resp)       to_initiator_1          <- mkFIFO;
  
   // TARGET FIFOs

   // ---- FIFOs for outgoing requests (to targets)
   FIFO#(Socket_Req)        to_target_0             <- mkFIFO;
   FIFO#(Socket_Req)        to_target_1             <- mkFIFO;

   // ---- FIFOs for incoming resps (from targets)
   FIFO#(Socket_Resp)       from_target_0           <- mkFIFO;
   FIFO#(Socket_Resp)       from_target_1           <- mkFIFO;

   // ================================================================
   // RULES (the interconnect functionality)

   // ----------------------------------------------------------------
   // Requests from initiators to targets

   let req_from_initiator_0 = from_initiator_0.first();
  
   let req_from_initiator_1 = from_initiator_1.first();
  
   // ----------------
   // To Target 0

   // Initiator 0 to Target 0
   rule initiator_0_to_target_0
                   (addr_is_for_target_0 (req_from_initiator_0.reqAddr));
     to_target_0.enq (req_from_initiator_0);
     from_initiator_0.deq();
   endrule

   // Initiator 1 to Target 0
   rule initiator_1_to_target_0
                   (addr_is_for_target_0 (req_from_initiator_1.reqAddr));
     to_target_0.enq (req_from_initiator_1);
     from_initiator_1.deq();
   endrule

   // ----------------
   // To Target 1

   // Initiator 0 to Target 1
   rule initiator_0_to_target_1
                   (addr_is_for_target_1 (req_from_initiator_0.reqAddr));
      to_target_1.enq(req_from_initiator_0);
      from_initiator_0.deq();
   endrule

   // Initiator 1 to Target 1
   rule initiator_1_to_target_1
                   (addr_is_for_target_1 (req_from_initiator_1.reqAddr));
      to_target_1.enq(req_from_initiator_1);
      from_initiator_1.deq();
   endrule

   // ----------------------------------------------------------------
   // Responses from targets to initiators

   let resp_from_target_0 = from_target_0.first();

   let resp_from_target_1 = from_target_1.first();

   // ----------------
   // To Initiator 0

   // Target 0 to Initiator 0
   rule target_0_to_initiator_0
                   (addr_is_for_initiator_0 (resp_from_target_0.respAddr));
     to_initiator_0.enq (resp_from_target_0);
     from_target_0.deq();
   endrule

   // Target 1 to Initiator 0
   rule target_1_to_initiator_0
                   (addr_is_for_initiator_0 (resp_from_target_1.respAddr));
     to_initiator_0.enq (resp_from_target_1);
     from_target_1.deq();
   endrule

   // ----------------
   // To Initiator 1

   // Target 0 to Initiator 1
   rule target_0_to_initiator_1
                   (addr_is_for_initiator_1 (resp_from_target_0.respAddr));
     to_initiator_1.enq (resp_from_target_0);
     from_target_0.deq();
   endrule

   // Target 1 to Initiator 1
   rule target_1_to_initiator_1
                   (addr_is_for_initiator_1 (resp_from_target_1.respAddr));
     to_initiator_1.enq (resp_from_target_1);
     from_target_1.deq();
   endrule

   // ================================================================
   // INTERFACES

   // ---- Initiators
   interface initiator_0 = fifos_to_slave_ifc (from_initiator_0, to_initiator_0);
   interface initiator_1 = fifos_to_slave_ifc (from_initiator_1, to_initiator_1);

   // ---- Targets
   interface target_0    = fifos_to_master_ifc (to_target_0, from_target_0);
   interface target_1    = fifos_to_master_ifc (to_target_1, from_target_1);

endmodule: mkSwitch

endpackage: Switch
