package Tb;

// ================================================================
// Copyright (c) Bluespec, Inc., 2007 All Rights Reserved

// ----------------------------------------------------------------
// A simple testbench for the Switch

// ----------------
// Imports from BSV libs

import LFSR:: *;
import FIFO:: *;
import GetPut:: *;
import ClientServer:: *;
import Connectable:: *;

// ----------------
// Imports for the design

import Socket_IFC::*;
import Switch::*;

// ----------------------------------------------------------------

(* synthesize *)
module mkTb (Empty);

   // Instantiate initiators and targets;
   Socket_master_ifc  initiator_0 <- mkInitiatorModel (0);
   Socket_master_ifc  initiator_1 <- mkInitiatorModel (1);

   Socket_slave_ifc   target_0    <- mkTargetModel (0);
   Socket_slave_ifc   target_1    <- mkTargetModel (1);

   // Instantiate the switch
   Switch_IFC the_switch <- mkSwitch ();

   // Make the connections
   mkConnection (initiator_0, the_switch.initiator_0);
   mkConnection (initiator_1, the_switch.initiator_1);
   mkConnection (target_0,    the_switch.target_0);
   mkConnection (target_1,    the_switch.target_1);
endmodule: mkTb

// ----------------------------------------------------------------
// Initiator module

module mkInitiatorModel #(Integer id)
                        (Socket_master_ifc);

   // Random number generator for requests
   // Use a different pseudo-random number generator seed for each instance
   // so that they generate different sequences
   Bit#(32) feed = case (id)
                       0 : return 32'h80000057;
                       1 : return 32'h80000062;
                       2 : return 32'h8000007A;
                       3 : return 32'h80000092;
                       4 : return 32'h800000B9;
                       5 : return 32'h800000BA;
                       6 : return 32'h80000106;
                       7 : return 32'h80000114;
                       8 : return 32'h8000012D;
                       9 : return 32'h8000014E;
                      10 : return 32'h8000016C;
                      11 : return 32'h8000019F;
                      12 : return 32'h800001A6;
                      13 : return 32'h800001F3;
                      14 : return 32'h8000020F;
                      default : return 32'h800002CC;
                   endcase;

   LFSR#(Bit#(32)) rand32 <- mkFeedLFSR(feed);

   // ----------------

   FIFO#(Socket_Req)   reqs   <- mkFIFO;    // Outgoing requests
   FIFO#(Socket_Resp)  resps  <- mkFIFO;    // Incoming responses

   FIFO#(Socket_Resp)  expected_resps_0 <- mkSizedFIFO (10); // Expected responses to RD reqs to 0
   FIFO#(Socket_Resp)  expected_resps_1 <- mkSizedFIFO (10); // Expected responses to RD reqs to 1

   // ----------------
   // Generating requests

   rule gen_reqs;
      let randx   = rand32.value();
      ReqOp op   = (((randx & 7) > 5) ? WR : RD);
      let   addr = zeroExtend (randx[15:4]);
      let   data = zeroExtend (randx[31:16]);
      let req = Socket_Req {reqOp: op,
                            reqInfo: fromInteger (id),
                            reqAddr: addr,
                            reqData: data};
      reqs.enq (req);
      rand32.next();

      if (op == RD) begin
         if (req.reqAddr[0] == 0) begin
            let resp = Socket_Resp {
                          respOp: OK,
                          respInfo: 0,
                          respAddr: fromInteger (id),
                          respData: req.reqData
                      };
            expected_resps_0.enq (resp);
         end
         else begin
            let resp = Socket_Resp {
                          respOp: OK,
                          respInfo: 1,
                          respAddr: fromInteger (id),
                          respData: req.reqData
                      };
            expected_resps_1.enq (resp);
         end
      end

      $write ("Tb: initiator %d sent request ", id);
      display_Socket_Req (req);
  endrule

   // ----------------
   // Fielding responses

   let resp = resps.first ();

   rule accept_resps_from_0 (resp.respInfo == 0);
     resps.deq ();

     $write ("Tb: initiator %d received response ", id);
     display_Socket_Resp (resp);
     let resp1 = expected_resps_0.first ();
     expected_resps_0.deq();
     if (resp1 != resp) begin
         $write ("ERROR: does not match expected response ");
         display_Socket_Resp (resp1);
         $finish (1);
     end
     else
         $display ("    Response is ok");
   endrule

   rule accept_resps_from_1 (resp.respInfo != 0);
     resps.deq ();

     $write ("Tb: initiator %d received response ", id);
     display_Socket_Resp (resp);
     let resp1 = expected_resps_1.first ();
     expected_resps_1.deq();
     if (resp1 != resp) begin
         $write ("ERROR: does not match expected response ");
         display_Socket_Resp (resp1);
         $finish (1);
     end
     else
         $display ("    Response is ok");
   endrule

   return  fifos_to_master_ifc (reqs, resps);

endmodule

// ----------------------------------------------------------------
// Targets

module mkTargetModel #(Integer id)
                     (Socket_slave_ifc);

   FIFO#(Socket_Req)  reqs  <- mkFIFO;
   FIFO#(Socket_Resp) resps <- mkFIFO;

   // ----------------
   // Responding to requests

   rule respond;
      let req = reqs.first ();
      reqs.deq();
      $write ("Tb: target %d received request ", id);
      display_Socket_Req (req);
      if (req.reqOp == RD) begin
         let resp = Socket_Resp {
                       respOp: OK,
                       respInfo: fromInteger (id),
                       respAddr: (req.reqInfo[0]),
                       respData: req.reqData
                   };
         resps.enq (resp);
         $write ("              Sent response ");
         display_Socket_Resp (resp);
      end
   endrule

   return  fifos_to_slave_ifc (reqs, resps);

endmodule

// ================================================================

endpackage: Tb
