package DMA ;

// ================================================================
// Copyright (c) Bluespec, Inc., 2007 All Rights Reserved


import Vector :: * ;
import Socket_IFC :: * ;
import FIFOS :: * ;
import EdgeFIFOs :: * ;


//  Here is the interface definition for the DMA.  It depends on definitions
//   in Socket_IFC.bsv.
interface DMA ;
   interface Socket_slave_ifc    cfg ; // interface to configuration port
   interface Socket_master_ifc   mmu ; // interface to memory management unit
   interface Status_ifc          st;   // interrupts & status for all channels
endinterface  


typedef UInt#(16)  DMACounts ;
// The number of channels is a parameter.
typedef 4 NumChannels ;




(* doc = " This DMA has 4 channels, which allows it to process 4 transfers" *)
(* doc = " simultaneously. The main difference from a 1-channel DMA is that" *)
(* doc = " we need to create vectors (arrays) of each channel's resources." *)
(* doc = " Each channel will have a unique set of registers to track its" *)
(* doc = " state." *)

(* synthesize *)
module mkDMA( DMA );    



   // For each socket, we need 2 fifos, one for requests, and one
   // for responses.  These fifos connect to the
   // interface sockets for the DMA module.

   ////////////////////////////////////////////////////////////////   
   // The fifos for the config port -- these are 1 element pipeline fifos.
   (* doc = "a fifo to hold incoming configuration requests" *)
   FIFOF#(Socket_Req)   cnfReqF  <- mkPipelineFIFO_ug_enq ;
   (* doc = "a fifo to hold outgoing configuration responses" *)
   FIFOF#(Socket_Resp)  cnfRespF <- mkBypassFIFO_ug_deq ;


   ////////////////////////////////////////////////////////////////   
   // The fifos for the MMU port
   (* doc = "a fifo to hold outgoing requests towards the memory" *)
   FIFOF#(Socket_Req)  mmuReqF  <- mkPipelineFIFO_ug_deq ;
   (* doc = "a fifo to hold incoming responses from the memory" *)
   FIFOF#(Socket_Resp) mmuRespF <- mkBypassFIFO_ug_enq ;
   


   ////////////////////////////////////////////////////////////////   
   // We need some registers to control the DMA transfers.
   // Each channel will have its own set of unique registers.  Each
   // type of register will be instantiated as a vector of registers, where
   // there is one for each channel.  The size of the vectors is the 
   // number of channels (NumChannels).  Access to the registers will be done
   // simply by using the index for the channel (e.g. 0 for channel 0; 1 for
   // channel 1; etc...).

   // Here we declare the channel registers; 
   // the instantiations are in the "for" loop below
   Vector#(NumChannels,Reg#(Bool)) dmaEnabledRs = newVector;
   Vector#(NumChannels,Reg#(ReqAddr)) readAddrRs = newVector;
   Vector#(NumChannels,Reg#(DMACounts)) readCntrRs = newVector;
   Vector#(NumChannels,Reg#(DMACounts)) currentReadRs = newVector;
   Vector#(NumChannels,Reg#(DMACounts)) currentWriteRs = newVector;
   Vector#(NumChannels,Reg#(ReqAddr)) destAddrRs = newVector;
   Vector#(NumChannels,FIFO#(ReqData)) responseDataFs = newVector;
   Vector#(NumChannels,FIFO#(ReqAddr)) destAddrFs = newVector;
   for (Integer i=0; i<valueof(NumChannels); i=i+1)
   begin
       //  dmaEnabledRs[i] is to set to turn a transfer on for
       //   that channel.  It is set back to 0 when the transfer is complete.
       (* doc = "This register indicates whether a transfer is enabled in the channel." *)
       	dmaEnabledRs[i] <- mkReg( False ) ;
       
       //  readAddrRs[i] is the current address from which to
       //   read the next data on that channel.
       (* doc = "A register to hold the read address (1 per channel)" *)
       	readAddrRs[i] <- mkReg(0) ;
       
       //  readCntrRs[i] holds the total number of data words to be 
       //   transferred for that channel.
       (* doc = "A register for the number of data-words to be transferred (1 per channel)" *)
       	readCntrRs[i] <- mkReg(0) ;
       
       //  currentReadRs[i] holds the total number of words read
       //   for that channel.
       (* doc = "A counter to hold number of words already read (1 per channel)" *)
       	currentReadRs[i] <- mkReg(0) ;
       
       //  currentWriteRs[i] holds the total number of words written for
       //   that channel.
       (* doc = "A counter to hold number of words already written (1 per channel)" *)
       	currentWriteRs[i] <- mkReg(0) ;
       
       //  destAddrRs[i] holds to current destination address for the next
       //   transfer in the DMA for that channel.  This is incremented
       //   throughout each DMA.
       (* doc = "The destination address" *)
       	destAddrRs[i]     <- mkReg(0) ;
            
       //  This internal FIFO is used to pass the read response to 
       //   the write side.  There is one for each channel.
       (* doc = " a FIFO to pass the read response to the write side" *)
       (* doc = " thus allowing pending transations and concurrency" *)
       	responseDataFs[i] <- mkFifo2;
       
       //  This internal FIFO is used to enqueue the destination
       //   address for the next data transfer.  There is one for each channel.
       (* doc = " We also want to pass the destination address for each read over" *)
       (* doc = " to the write side" *)
       (* doc = " The depth of this fifo limits the number of outstanding reads" *)
       (* doc = " which may be pending before the write.  The maximun outstanding" *)
       (* doc = " reads depends on the overall latency of the read requests." *)
       	destAddrFs[i] <- mkFifo4 ;
    end // for loop




   /////////////////////////////////////////////////////////////////
   ///  DMA rules //////////////////////////////////////////////////
   /////////////////////////////////////////////////////////////////
   //
   // We define a function inside the module so it can create the rules
   // for each channel in a single call.
   // The function takes as arguments the fifos and the relevant channel
   // number.  It returns a set of rules for that channel.  These rules
   // describe much of the internal module behavior for the DMA.
   //
   function Rules generatePortDMARules (FIFOF#(Socket_Req)  requestF,
                                        FIFOF#(Socket_Resp) responseF,
                                        Integer channelNum
                                        );
      return
      rules

         // The following attribute defines the relative priority (urgency)
         // for the rules created with this function.  This essentially
         // defines the priority for rules firing in the same cycle, in the
         // event that multiple channel rules want to fire, but they conflict.
         // In this statement, 'startWrite' is highest priority.
         (* descending_urgency = "startWrite, startRead" *)


         (* doc = " This rule starts a write process" *)
         (* doc = " Note that this rule conflicts with rule " *)
         (* doc = " startRead, so we make this rule more urgent." *)
         (* doc = " I.e., finish writing before you start" *)
         (* doc = " reading more." *)
         rule startWrite ;
            // Generate a Write 
            let wreq = Socket_Req {reqAddr : destAddrFs[channelNum].first(),
                                   reqData : responseDataFs[channelNum].first(),
                                   reqOp   : WR,
                                   reqInfo : fromInteger(1 + 2*channelNum) }; 

            // enqueue the request.
            requestF.enq( wreq ) ;
      
            // Some other house keeping -- removing the data from the fifos
            destAddrFs[channelNum].deq() ;
            responseDataFs[channelNum].deq() ;
         endrule


         (* doc = " This rule waits for the write to finish" *)
         rule finishWrite ((responseF.first().respInfo == fromInteger(1 + 2*channelNum)) ) ;
            responseF.deq() ;          // take the response data and finish
            currentWriteRs[channelNum] <=currentWriteRs[channelNum] + 1;
         endrule


         (* doc = " This starts a read, when the dma is enabled and there" *)
         (* doc = " are data to move (and we are in the right state)" *)
         rule startRead (dmaEnabledRs[channelNum] && 
                         readCntrRs[channelNum] > currentReadRs[channelNum] ) ;
      
            // Create a read request, and enqueue it.
            // Since there can be multiple pending requests, either read or
            // writes, we use the reqInfo field to mark these.  reqInfo
            // indicates the channel and whether it is a RD or WR.  E.g.:
	    //   0 means RD for channel 0
            //   1 means WR for channel 0
            //   2 means RD for channel 1
            //   etc...
            let req = Socket_Req {reqAddr : readAddrRs[channelNum],
                                  reqData : 0,
                                  reqOp   : RD,
                                  reqInfo : fromInteger(0 + 2*channelNum)};      
            requestF.enq( req ) ;
      
            // Enqueue the Write destination address
            destAddrFs[channelNum].enq( destAddrRs[channelNum] ) ;

            // Some house keeping -- increment the read address,
            // the number of words read and the destination address.
            readAddrRs[channelNum] <= readAddrRs[channelNum] + 1 ;
            currentReadRs[channelNum] <= currentReadRs[channelNum] + 1 ;
            destAddrRs[channelNum] <= destAddrRs[channelNum] + 1 ;
         endrule



         (* doc = " We finish a read when we see the correct respInfo on the mmu" *)
         (* doc = " response fifo." *)
         rule finishRead (responseF.first().respInfo == fromInteger(0 + 2*channelNum) ) ;
            // grab the data from the mmu reponse fifo
            Socket_Resp resp = responseF.first() ;     
            responseF.deq() ;

            // Note: this spec/design assumes responses don't fail (e.g. we
            //  don't get an error in the response or the request doesn't
            //  fail).
            
            // Pass the read data to the write "side" of the dma
            responseDataFs[channelNum].enq( resp.respData );
         endrule




      endrules ;
   endfunction   // End of function generatePortDMARules()








   // This function generates the rules for each channel to do the clean up
   //  required whenever a transfer on a channel is complete.

   function Rules generateTransferDoneRules( Integer channelNum );
      return
      rules
         (* doc = " Conditions to mark when transfer is done" *)
         rule markTransferDone (dmaEnabledRs[channelNum] &&
                                (currentWriteRs[channelNum] == readCntrRs[channelNum]) &&
                                (currentReadRs[channelNum]  == readCntrRs[channelNum]) ) ;
            dmaEnabledRs[channelNum] <=False ; 
            currentWriteRs[channelNum] <=0 ;
            currentReadRs[channelNum] <=0 ;
         endrule
      endrules ;
   endfunction


   




   // The following section generates the rules.
   //  Then, we can be manipulate them further.

   // Generate the PortDMA rules for channel 0
   Rules r01 = generatePortDMARules( mmuReqF, mmuRespF, 0 ) ;
   // Generate the PortDMA rules for channel 1
   Rules r11 = generatePortDMARules( mmuReqF, mmuRespF, 1 ) ;
   // Generate the PortDMA rules for channel 2
   Rules r21 = generatePortDMARules( mmuReqF, mmuRespF, 2 ) ;
   // Generate the PortDMA rules for channel 3
   Rules r31 = generatePortDMARules( mmuReqF, mmuRespF, 3 ) ;
   // Generate the transfer done rules for channel 0
   Rules td0 = generateTransferDoneRules(0) ;
   // Generate the transfer done rules for channel 1
   Rules td1 = generateTransferDoneRules(1) ;
   // Generate the transfer done rules for channel 2
   Rules td2 = generateTransferDoneRules(2) ;
   // Generate the transfer done rules for channel 3
   Rules td3 = generateTransferDoneRules(3) ;
   





   // The set of rules above have conflicts.  E.g. two startRead's
   // conflict.  The rJoinDescendingUrgency essentially establishes
   // the relative priority of two rules.  The first rule is the higher
   // priority.  rJoinDescendingUrgency also joins the two rules into
   // one entity with the combined rules executing/firing with their
   // relative priorities.
   // 
   // e.g. in this next line, r01 is given higher priority than r11
   // (we are saying channel 0 gets higher priority than channel 1)
   Rules r1 = rJoinDescendingUrgency( r01, r11) ;
   Rules r2 = rJoinDescendingUrgency( r21, r31) ;
   

   // The following lines successively pull rules into a single entity
   //  while at the same time establishing the following priority:
   //  r01 > r11 > r21 > r31 > td0 > td1 > td2 > td3
   r2 = rJoinDescendingUrgency( r1, r2 ) ;
   r2 = rJoinDescendingUrgency( r2, td0 ) ;
   r2 = rJoinDescendingUrgency( r2, td1 ) ;
   r2 = rJoinDescendingUrgency( r2, td2 ) ;
   r2 = rJoinDescendingUrgency( r2, td3 ) ;

   // This line just takes the rule descriptions and make them part
   //  of the module behavior.
   addRules( r2 ) ;

         




   //////////////////////////////////////////////////////////
   ///                                                    ///
   ///  Rules and other code to support DMA configuration ///
   ///                                                    ///
   //////////////////////////////////////////////////////////

   // Add a zero-size register as a default for invalid addresses
   Reg#(Bit#(0)) nullReg <- mkReg( ? ) ;

   // For ease of development we want all registers to look like 32
   // bit registers -- the data size of the config socket.
   // The following is a function to map from address to specific registers.
   // The 12 bit address is split into two fields:
   // 4 bits to select the channel AND 8 to select the register.
   //
   // This function returns the selected register, by channel and type, as
   // specified by the addr.
   
   function Reg#(ReqData) selectReg( ReqAddr addr ) ;
      Bit#(8) regSel = truncate( addr ) ;
      Bit#(4) channelSel = truncate ( addr >> 8 ) ;
      return
      case ( regSel )
         8'h00 :  return regAToRegBitN( readAddrRs[channelSel] ) ;
         8'h04 :  return regAToRegBitN( readCntrRs[channelSel] ) ;
         8'h08 :  return regAToRegBitN( destAddrRs[channelSel] ) ;
         8'h0C :  return regAToRegBitN( dmaEnabledRs[channelSel] ) ;         
         default:  return regAToRegBitN( nullReg ) ;
      endcase ;
   endfunction


   (* doc = " A rule for writing to a configuration register and" *)
   (* doc = " and generating a response." *)
   rule writeConfig ( cnfReqF.first().reqOp == WR ) ;
      let req =  cnfReqF.first() ;
      cnfReqF.deq() ;

      // Select and write the register 
      let thisReg = selectReg( req.reqAddr ) ;
      thisReg <= req.reqData ;

      // Now generate the response and enqueue
      let resp = Socket_Resp {respOp   : OK,
                              respAddr : 0,
                              respInfo : req.reqInfo,
                              respData : req.reqData } ;
      cnfRespF.enq( resp ) ;
   endrule
   

   (* doc = " A rule for reading a configuration register" *)
   (* doc = " and generating a response." *)
   rule readConfig ( cnfReqF.first().reqOp == RD ) ;
      let req =  cnfReqF.first() ;
      cnfReqF.deq() ;

      // Select the register 
      let thisReg = selectReg( req.reqAddr ) ;

      // Now generate the response and enqueue
      let resp = Socket_Resp {respOp   : OK,
                              respAddr : 0,
                              respInfo : req.reqInfo,
                              respData : thisReg } ;
      cnfRespF.enq( resp ) ;
   endrule


   //  Define the priority for the config update rules.  Highest
   //   priority is the 'writeConfig' rule.
   (* descending_urgency = 
    "writeConfig, readConfig, unknownConfig"  *)
   (* doc = " this is basically what's left for NOP command" *)
   rule unknownConfig ( True ) ; 
      let req =  cnfReqF.first() ;
      cnfReqF.deq() ;

      // Select the register 
      let thisReg = selectReg( req.reqAddr ) ;

      // Now generate the response and enqueue
      let resp = Socket_Resp {respOp   : NOP,
                              respAddr : 0,
                              respInfo : req.reqInfo,
                              respData : thisReg } ;
      cnfRespF.enq( resp ) ;
   endrule
   




   ////////////////////////////////////////////////////////////////  
   ////////////////////////////////////////////////////////////////  
   // 
   // Create the interfaces by connecting the fifo interfaces to the
   //  the socket ports.

   interface cfg  = fifos_to_slave_ifc(  cnfReqF, cnfRespF ) ;

   interface mmu = fifos_to_master_ifc( mmuReqF, mmuRespF ) ;

   interface Status_ifc st;

       // An interrupt per channel, to enable the channel
       method Action ch0_int;
           dmaEnabledRs[0] <= True;
       endmethod
       method Action ch1_int;
           dmaEnabledRs[1] <= True;
       endmethod
       method Action ch2_int;
           dmaEnabledRs[2] <= True;
       endmethod
       method Action ch3_int;
           dmaEnabledRs[3] <= True;
       endmethod

       // An output port per channel indicating its current status
       method Bool  ch0_busy = dmaEnabledRs[0];
       method Bool  ch1_busy = dmaEnabledRs[1];
       method Bool  ch2_busy = dmaEnabledRs[2];
       method Bool  ch3_busy = dmaEnabledRs[3];
   endinterface

endmodule


endpackage


         
