import Vector ::*;

// ================================================================
// Copyright (c) Bluespec, Inc., 2007 All Rights Reserved

// this is an interface to a plain fifo.
interface FIFO #(type a);
    method Action enq(a x1);
    method Action deq();
    method a first();
    method Action clear();
endinterface: FIFO

// this is an interface to a sized fifo, where the depth 
// is a parameter to the module.
interface FIFO_ #(numeric type depth, type a);
    method Action enq(a x1);
    method Action deq();
    method a first();
    method Action clear();
endinterface: FIFO_

// this is a more detaled version of a FIFO interface,
// where the "notFull" and "notEmpty" signals are exposed
interface FIFOF #(type a);
    method Action enq(a x1);
    method Action deq();
    method a first();
    method Bool notFull();
    method Bool notEmpty();
    method Action clear();
endinterface: FIFOF

// a module definition for a sized FIFO
module mkSizedFIFO(FIFO_#(depth, a))
    provisos (Bits#(a, sa), Log#(depth, addr_size), Literal#(a));

    // instantiating an array of registers
    Vector#(depth, Reg#(a)) state = newVector;
    for (Integer i = 0; i < valueOf(depth); i = i+1)
        state[i] <- mkReg(0);
    // a register for the write pointer
    Reg#(UInt#(addr_size)) wrPtr <- mkReg(0);
    // a register for the read pointer
    Reg#(UInt#(addr_size)) rdPtr <- mkReg(0);
    // if the fifo is full, but a user is trying to dequeue from the fifo,
    // we want to allow simultanious enq. For that - we'll put a pulsewire
    // to indicate if the deq method was called
    PulseWire deq_called <- mkPulseWire;
    // when calling "clear" and "enq" at the same time - we would like the "clear"
    // method to get a higher priority, and we'd like to block the "enq" method. 
    // This pulseWire will make the call to the clear method visible
    PulseWire clr_called <- mkPulseWire;

    // local variables, for convenience
    Bool notFull  = (wrPtr - rdPtr != fromInteger(valueof(depth) - 1));
    Bool notEmpty = (rdPtr != wrPtr);

    // interface definition
    // enq can be called when the fifo is not full, or if there is a simultanious deq
    method Action enq(a x1) if ( !clr_called && (notFull || deq_called)); //we want to block enq when clearing
        state[wrPtr] <= x1;
        wrPtr        <= (wrPtr == (fromInteger(valueOf(depth)-1))) ? 0 : wrPtr+1;
    endmethod
    // deq can be called as long as the FIFO is not empty
    method Action deq() if (notEmpty);
        rdPtr        <= (rdPtr == (fromInteger(valueOf(depth)-1))) ? 0 : rdPtr+1;
        deq_called.send(); // now it will be visible on the enq side
    endmethod
    // the first value can be read as long as the FIFO is not empty
    method a first() if (notEmpty);
        return state[rdPtr];
    endmethod
    // the clear method can always be called
    method Action clear();
        for (Integer i = 0; i < valueof(depth); i = i+1)
            (state[i]) <= 0;
            clr_called.send();
    endmethod
endmodule
