// This package contains a testbench for the DMA package.
package TestBench;

import FIFOS :: * ;
import StmtFSM :: * ;
import EdgeFIFOs :: * ;
import Socket_IFC :: * ;

import DMA :: * ;
import Targets :: * ;

(* doc = " Create a testbench module, which consists of a DMA and a simple" *)
(* doc = " MMU (slave).  The testbench configures the DMA and then waits for the" *)
(* doc = " DMA transfer to finish." *)

// Testbench interface is empty, thus having just a clock and reset
(* synthesize *)
module sysTestBench () ;

   (* doc = " Instantiate the DMA - the DUT" *)
   DMA dma <- mkDMA ;
   
   (* doc = " In the test bench we want to synthesize a memory" *)
   Socket_slave_ifc  mem1 <- mkDummyMemory( "memA" ) ;

   // Now connect the DMA to the memory
   Empty mmuConnection1 <- mkConnection( mem1, dma.mmu ) ;

   // For ease, we create a master socket by
   // instantiating 2 fifos and connecting them to the DMA config
   // slave socket.  The testbench will be able to send requests and
   // receive responses to theses fifos
   (* doc = "a fifo to store requests to the config port" *)
   FIFOF#(Socket_Req)   cnf_req_f  <- mkPipelineFIFO_ug_deq ;
   (* doc = "a fifo to store responses from the config port" *)
   FIFOF#(Socket_Resp)  cnf_resp_f <- mkPipelineFIFO_ug_enq ;
   let  configMaster = fifos_to_master_ifc( cnf_req_f, cnf_resp_f ) ;
   // and connect to the DMA
   Empty cfgConnection <- mkConnection(configMaster, dma.cfg ) ;
   
   
   // For convenience, we define a function to build the request
   // structure for configuring the DMA
   function Action genConfig (ReqAddr maddr, 
                              ReqOp   cmd, 
                              ReqData mdata ) ;      
      let req = Socket_Req {reqAddr : maddr,
                            reqData : mdata,
                            reqOp   : cmd ,
                            reqInfo : 14 } ;
      return 
      action
         cnf_req_f.enq( req ) ;
      endaction;
   endfunction

   Stmt dmaConfig = 
   seq
      $display( "dmaConfig starting" ) ;
      // dmaConfig is a simple FSM to configure the DMA
      // Move 20 items from hex address 1000 to 5000 
      // Note that each request requires a deq from the fifo to take
      // the reponse from the DMA.
      genConfig( 'h00, WR, 'h0000_1000 ) ;
      genConfig( 'h04, WR,  20 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h08, WR, 'h0000_5000 ) ;
      cnf_resp_f.deq() ;
      
      // Move 20 items from hex address 2000 to 7000 
      // Note that each request requires a deq from the fifo to take
      // the reponse from the DMA.
      genConfig( 'h100, WR, 'h0000_2000 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h104, WR,  20 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h108, WR, 'h0000_7000 ) ;
      cnf_resp_f.deq() ;
      
      // Move 20 items from hex address 3000 to 4000 
      // Note that each request requires a deq from the fifo to take
      // the reponse from the DMA.
      genConfig( 'h200, WR, 'h0000_3000 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h204, WR,  20 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h208, WR, 'h0000_4000 ) ;
      cnf_resp_f.deq() ;
            
      // Move 20 items from hex address 4000 to 6000 
      // Note that each request requires a deq from the fifo to take
      // the reponse from the DMA.
      genConfig( 'h300, WR, 'h0000_4000 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h304, WR,  20 ) ;
      cnf_resp_f.deq() ;
      genConfig( 'h308, WR, 'h0000_6000 ) ;
      cnf_resp_f.deq() ;

      action
         cnf_resp_f.deq() ;
         $display( "DMAConfig done" ) ;
      endaction
   endseq ;
   // Another FSM to control overall sequeining
   Stmt testCtrl =
   seq
      dmaConfig ;
      $display( "Starting the DMA controller channel 0" ) ;
      dma.st.ch0_int();
         await( !dma.st.ch0_busy() );
      $display( "Channel 0 is complete." ) ;

      $display( "Finished  run of channel 0, starting channel 1" ) ;
      dma.st.ch1_int();
         await( !dma.st.ch1_busy() );
      $display( "Channel 1 is complete." ) ;

      $display( "Finished  run of channel 1, starting channel 2" ) ;
      dma.st.ch2_int();
         await( !dma.st.ch2_busy() );
      $display( "Channel 2 is complete." ) ;

      $display( "Finished  run of channel 2, starting channel 3" ) ;
      dma.st.ch3_int();
         await( !dma.st.ch3_busy() );
      $display( "Channel 3 is complete." ) ;

      $display( "Finished  run of channel 3, starting all channels" ) ;
      action
         dma.st.ch0_int();
         dma.st.ch1_int();
         dma.st.ch2_int();
         dma.st.ch3_int();
      endaction
      $display( "All channels started" ) ;
      par
         seq
            await( !dma.st.ch0_busy() );
            $display( "Channel 0 is complete." ) ;
         endseq
         seq
            await( !dma.st.ch1_busy() );
            $display( "Channel 1 is complete." ) ;
         endseq
         seq
            await( !dma.st.ch2_busy() );
            $display( "Channel 2 is complete." ) ;
         endseq
         seq
            await( !dma.st.ch3_busy() );
            $display( "Channel 3 is complete." ) ;
         endseq
      endpar
      $display( "All channels finished" ) ;
      
   endseq ;
   
   // Create an FSM from the test sequence.
   mkAutoFSM( testCtrl ) ;
   
   // 
   Reg#(int) c <- mkReg(0) ;
   rule ticker ( True ) ;
      c <= c + 1;
      $display("cycle: %0d", c ) ;
      if ( c > 5000 ) 
         begin
            $display ("Simulation finished because of timeout" ) ; 
            $finish(0);
         end
   endrule
   
endmodule


endpackage 
